/****
    coval version 0.5
    (C) 2011.02.06 Takeshi Ogihara
 ****/
#include <limits.h>
#include "coval.h"

#if USE_LONGDOUBLE
#define  _LDOUBLE	long double
#else
#define  _LDOUBLE	double
#endif

#if USE_LONGLONG
#define  _LLONG		long long
#define  _ULLONG	unsigned long long
#else
#define  _LLONG		long
#define  _ULLONG	unsigned long
#endif

#define Coval_bridge_ID		(coval_callback)(1)	// special value

enum { T_NONE,
	T_CHR, T_UCHR, T_SHT, T_USHT, T_INT, T_UINT, T_LNG, T_ULNG,
#if USE_LONGLONG
	T_LLNG, T_ULLNG,
#endif
	T_FLT, T_DBL,
#if USE_LONGDOUBLE
	T_LDBL,
#endif
};

#define _makeDiffData(A)  *(char *)&(A)->_d = *(char *)(A)->_p + 1

void _bindCoval(coval_base *target, coval_base *cov, void *pp)
{
	coval_base *center;
	if (target == NULL || cov == NULL || target == cov || cov->_p != NULL)
		return;
	if (target->_p == NULL) { // first bind
		center = target;
		center->_p = &(center->_d); // make center coval
		center->_b._cinfo._count = 1;
		center->_b._cinfo._properties = 0; // default values
		center->_nx = center->_pv = NULL;
	}else {
		center = IsCenter(target) ? target : target->_b._cent;
		center->_b._cinfo._count++;
	}
	cov->_p = (pp != NULL) ? pp : center->_p; // pp points a certain member
	cov->_b._cent = center;
	if (cov->_cb) { /* make a ring */
		if (center->_nx) {
			coval_base *p, *q;
			cov->_nx = q = center->_nx;
			cov->_pv = p = q->_pv;
			p->_nx = cov;
			q->_pv = cov;
			if (center->_pv == p) center->_pv = cov;
		}else {
			center->_nx = cov;
			center->_pv = NULL;
			cov->_nx = cov->_pv = cov;
		}
		_makeDiffData(cov);
		// Bridge: make different value in the bridge
	}
}

void _set_coval_function(coval_base *cov, coval_callback fn, void *arg)
{
/*	if (cov->_p == NULL) { // checked in MACRO SetFunctionOfCoval
		cov->_cb = fn;  cov->_ag = arg;
	}
 */
	coval_base *center;
	int ringflag = (((cov->_cb == NULL) ^ (fn == NULL)) == 1);
	/* lock(cov->_mtx); should be in a critical section */
	cov->_cb = fn;
	cov->_ag = arg;
	/* unlock(cov->_mtx); should be in a critical section */
	if (IsCenter(cov) || ringflag == 0)
		return;
	center = cov->_b._cent;
	if (fn) { // cov->_cb == NULL && fn
		_bindCoval(center, cov, NULL);
		center->_b._cinfo._count--;	// cov is registered double
	}else { // cov->_cb && fn == NULL
		_unbindCoval(cov);
		_bindCoval(center, cov, NULL);
	}
}

void _unbindCoval(coval_base *cov)
{
	// This part is done in MACRO
	// if (cov->_p == NULL || IsCenter(cov))
	//	return;	/* center coval can't unbind self */

	coval_base *center = cov->_b._cent;

	if (cov->_cb) { /* cov is in the ring */
		coval_base *p, *q;
		p = q = center->_nx;
		if (p == cov && cov->_nx == cov) { // only one
			center->_nx = center->_pv = NULL;
		}else {
			do {
				if (p == cov) {
					p = cov->_pv;
					q = cov->_nx;
					p->_nx = q;
					q->_pv = p;
					if (center->_nx == cov)
						center->_nx = q;
					if (center->_pv == cov)
						center->_pv = NULL;
					break;
				}
			} while ((p = p->_nx) != q);
		}
		cov->_nx = cov->_pv = NULL;
	}
	cov->_p = NULL;
	cov->_b._cent = NULL;
	if (--center->_b._cinfo._count == 0) // no member
		center->_p = NULL;
}

static void call_coval_bridge(coval_base *cov, bridge_base *brg)
{
	coval_base *tgt;
	unsigned int op, chk, from;
	int idx;
	_LLONG n;
	_ULLONG ul;
	_LDOUBLE f;
	int uflag;
	void *p, *q;

	idx = (cov == brg->_cv[0]) ? 0 : 1;
	if ((op = (brg->_sw[idx] & Func_mask)) == BR_NONE)
		return;	// Do nothing
	tgt = brg->_cv[1 - idx];
	if (tgt->_p == NULL)
		return; /* not connected */
	chk = (brg->_sw[idx] & Check_mask);
	if (op == Func_mask) { // call conversion function
		int flag = (brg->_fn[idx].func)(tgt->_p, cov->_p);
		if (!chk || flag) 
			_coval_call(tgt);	/* as CovalAssign */
		return;
	}
	if (op == BR_CopyBytes) {
		unsigned char *pp = (unsigned char *)cov->_p;
		unsigned char *qq = (unsigned char *)tgt->_p;
		// pp != qq, because a bridge is not a center coval.
		unsigned int len = brg->_sz[idx];
		for (; len > 0; ++pp, ++qq, --len) {
			if (*pp != *qq) break;
		}
		if (len > 0) {
			while (len--)
				*qq++ = *pp++;
		}else if (chk)
			return;
		_coval_call(tgt);
		return;
	}

	/* typical conversions */
	uflag = 0;
	p = cov->_p;
	from = brg->_fn[idx].conv[0];
	switch (from) {
	case T_CHR: f = (_LDOUBLE)(n = *(char *)p); break;
	case T_SHT: f = (_LDOUBLE)(n = *(short *)p); break;
	case T_INT: f = (_LDOUBLE)(n = *(int *)p); break;
	case T_LNG: f = (_LDOUBLE)(n = *(long *)p); break;
	case T_LLNG: f = (_LDOUBLE)(n = *(_LLONG *)p); break;
	case T_FLT: n = (_LLONG)(f = *(float *)p); break;
	case T_DBL: n = (_LLONG)(f = *(double *)p); break;
#if USE_LONGDOUBLE
	case T_LDBL: n = (_LLONG)(f = *(_LDOUBLE *)p); break;
#endif
	case T_UCHR: f = (_LDOUBLE)(ul = *(unsigned char *)p);
		uflag = 1; break;
	case T_USHT: f = (_LDOUBLE)(ul = *(unsigned short *)p);
		uflag = 1; break;
	case T_UINT: f = (_LDOUBLE)(ul = *(unsigned int *)p);
		uflag = 1; break;
	case T_ULNG: f = (_LDOUBLE)(ul = *(unsigned long *)p);
		uflag = 1; break;
#if USE_LONGLONG
	case T_ULLNG: f = (_LDOUBLE)(ul = *(_ULLONG *)p);
		uflag = 1; break;
#endif
	}
	if (chk) {
		int flag;
		_LLONG nck;
		_ULLONG ulck;
		_LDOUBLE fck;
		void *r = (void *)&cov->_d;
		switch (from) {
		case T_CHR: flag = (n == (nck = *(char *)r)); break;
		case T_SHT: flag = (n == (nck = *(short *)r)); break;
		case T_INT: flag = (n == (nck = *(int *)r)); break;
		case T_LNG: flag = (n == (nck = *(long *)r)); break;
		case T_LLNG: flag = (n == (nck = *(_LLONG *)r)); break;
		case T_FLT: flag = (f == (fck = *(float *)r)); break;
		case T_DBL: flag = (f == (fck = *(double *)r)); break;
#if USE_LONGDOUBLE
		case T_LDBL: flag = (f == (fck = *(_LDOUBLE *)r)); break;
#endif
		case T_UCHR: flag = (ul == (ulck = *(unsigned char *)r));
			break;
		case T_USHT: flag = (ul == (ulck = *(unsigned short *)r));
			break;
		case T_UINT: flag = (ul == (ulck = *(unsigned int *)r));
			break;
		case T_ULNG: flag = (ul == (ulck = *(unsigned long *)r));
			break;
#if USE_LONGLONG
		case T_ULLNG: flag = (ul == (ulck = *(_ULLONG *)r));
			break;
#endif
		}
		if (flag)
			return; // do nothing
	}

	if (op == BR_NumNeg) {
		f = -f;
		if (uflag) {
			n = -1 * (_LLONG)ul;
			uflag = 0;
		}else
			n = -n;
	}else if (op == BR_LogicNeg) {
		n = uflag ? (ul == 0) : (n == 0);
		f = n;
	}

	q = tgt->_p;
	switch (brg->_fn[idx].conv[1]) {
	case T_CHR:
	case T_UCHR: *(char *)q = (char)(uflag ? ul : n); break;
	case T_SHT:
	case T_USHT: *(short *)q = (short)(uflag ? ul : n); break;
	case T_INT:
	case T_UINT: *(int *)q = (int)(uflag ? ul : n); break;
	case T_LNG:
	case T_ULNG: *(long *)q = (long)(uflag ? ul : n); break;
	case T_LLNG:
	case T_ULLNG: *(_LLONG *)q = (_LLONG)(uflag ? ul : n); break;
	case T_FLT: *(float *)q = (float)f; break;
	case T_DBL: *(double *)q = (double)f; break;
#if USE_LONGDOUBLE
	case T_LDBL: *(_LDOUBLE *)q = (_LDOUBLE)f; break;
#endif
	}
	/* as CovalAssign */
	_coval_call(tgt);
}

typedef struct {
	coval_base *fin, *currnet, *newcaller;
	int callerIsChecked;
} loopInfo;

void _coval_call(coval_base *caller)
{
	if (caller->_p == NULL)
		return;

	loopInfo info;
	int ret, breakable;
	coval_base *center, *p;

	center = IsCenter(caller) ? caller : caller->_b._cent;
	info.newcaller = caller;
	info.callerIsChecked = (caller->_cb == NULL);
	if (center->_nx == NULL) { // no ring
		if (center->_pv == NULL // not in loop
		&& center->_cb && (center->_callself || center != caller)) {
			info.fin = info.currnet = center;
			center->_pv = (coval_base *)&info; // flag on
			(void)center->_cb(center, center->_ag);
			center->_pv = NULL; // flag off
		}
		return;
	}

	if (center->_pv != NULL) { // modify during processing
		loopInfo *infp = (loopInfo *)center->_pv;
		if (infp->currnet != caller) {
			infp->newcaller = caller;
			infp->callerIsChecked = (caller->_cb == NULL);
		}
		return;
	}
	breakable = (center->_b._cinfo._properties & BIND_FLAG_BREAKABLE) != 0;
	center->_pv = (coval_base *)&info; // flag on
	info.fin = center->_nx;
	for (p = center->_nx; ; ) {
		if (p == center->_nx) {	// Special Case: call center's callback
			if (center->_cb
			&& (center->_callself || center != info.newcaller)) {
				info.currnet = center;
				ret = center->_cb(center, center->_ag);
				if (breakable && ret == COVAL_PROC_SUCCESS)
					break;
			}
			if (center == info.newcaller)
				info.callerIsChecked = 1;
		}
		info.currnet = p;
		if (p->_callself || p != info.newcaller) {
			if (p->_cb == Coval_bridge_ID) { /* Bridge */
				call_coval_bridge(p, p->_ag);
			}else {
				ret = p->_cb(p, p->_ag);
				if (breakable && ret == COVAL_PROC_SUCCESS)
					break;
			}
		}
		if (p == info.newcaller)
			info.callerIsChecked = 1;
		p = p->_nx;
		if (info.fin == p) info.fin = NULL;
		if (info.fin == NULL && info.callerIsChecked == 1)
			break;
	}
	center->_pv = NULL; // flag off
}

void _setBindingProperty(coval_base *cov, unsigned int prop, int flag)
{
	unsigned int b;
	if (cov->_p == NULL)
		return;	/* not bound */
	coval_base *center = IsCenter(cov) ? cov : cov->_b._cent;
	b = center->_b._cinfo._properties;
	if (flag) b |= prop;
	else b &= 0xff ^ prop;
	center->_b._cinfo._properties = b;
}

static int type_of_conv(unsigned int sz, unsigned int fmt)
{
	switch (fmt) {
	default:
	case BR_NONE: break;
	case BR_INT:
		if (sz == sizeof(char)) return T_CHR;
		if (sz == sizeof(short)) return T_SHT;
		if (sz == sizeof(int)) return T_INT;
		if (sz == sizeof(long)) return T_LNG;
#if USE_LONGLONG
		if (sz == sizeof(long long)) return T_LLNG;
#endif
		break;
	case BR_UINT: 
		if (sz == sizeof(unsigned char)) return T_UCHR;
		if (sz == sizeof(unsigned short)) return T_USHT;
		if (sz == sizeof(unsigned int)) return T_UINT;
		if (sz == sizeof(unsigned long)) return T_ULNG;
#if USE_LONGLONG
		if (sz == sizeof(unsigned long long)) return T_ULLNG;
#endif
	case BR_FLT:
		if (sz == sizeof(float)) return T_FLT;
		if (sz == sizeof(double)) return T_DBL;
#if USE_LONGDOUBLE
		if (sz == sizeof(long double)) return T_LDBL;
#endif
		break;
	}
	return T_NONE;
}

#define _CONV_TooLong	USHRT_MAX

void _coval_bridge_initialize(bridge_base *brg, unsigned sz1, unsigned sz2)
{
	coval_base *c1, *c2;
	unsigned int w;
	brg->_sz[0] = (sz1 >= _CONV_TooLong) ? _CONV_TooLong : sz1;
	brg->_sz[1] = (sz2 >= _CONV_TooLong) ? _CONV_TooLong : sz2;
	c1 = brg->_cv[0];
	c2 = brg->_cv[1];
	c1->_cb = c2->_cb = Coval_bridge_ID;
	c1->_ag = c2->_ag = (void *)brg;  // Note: Special case for Bridge.
	if (sz1 == sz2 && (w = type_of_conv(sz1, BR_UINT)) != T_NONE) {
		brg->_sw[0] = brg->_sw[1] = _BR_Conv; // deal as unsigned integer
		brg->_fn[0].conv[0] = brg->_fn[0].conv[1] = w;
		brg->_fn[1].conv[0] = brg->_fn[1].conv[1] = w;
	}else {
		brg->_fn[0].func = brg->_fn[1].func = (coval_bridge_convert)NULL;
		brg->_sw[0] = brg->_sw[1] = BR_CopyBytes; // copy data & no func
	}
}

int _setCovalBridgeConversion(bridge_base *brg, int idx, int conv, int check)
{
	int oth;
	unsigned int op, t_fr, t_to;
	unsigned char *prms;

	brg->_sw[idx] = check ? Check_mask : 0;
	if (conv == BR_NONE)
		return 1; // bridge does nothing
	op = (conv & _CONV_OpeMASK);
	brg->_sw[idx] |= op ? op : _BR_Conv;
	t_fr = conv & _CONV_FromMASK;
	t_to = (conv & _CONV_ToMASK) >> _CONV_ToShift;
	if (op == BR_LogicNeg && t_to == BR_FLT)
		return 0; // logical value should be integer
	prms = brg->_fn[idx].conv;
	oth = 1 - idx;
	if (op == BR_CopyBytes) {
		unsigned int sz, w;
		sz = brg->_sz[idx];
		if (sz != brg->_sz[oth] || sz == _CONV_TooLong)
			return 0; // ERROR
		w = type_of_conv(sz, BR_UINT);
		if (w == T_NONE)
			return 1; // BR_CopyBytes
		brg->_sw[idx] = _BR_Conv | (check ? Check_mask : 0);
		prms[0] = prms[1] = w;  // deal as unsigned integer
	}else { // _BR_Conv | BR_NumNeg | BR_LogicNeg
		prms[0] = type_of_conv(brg->_sz[idx], t_fr);
		prms[1] = type_of_conv(brg->_sz[oth], t_to);
		if (prms[1] == T_NONE || prms[2] == T_NONE) {
			brg->_sw[idx] = BR_CopyBytes;
			return 0; // ERROR
		}
	}
	return 1; // success
}
