/****
    coval version 0.5 
    (C) 2011.02.06 Takeshi Ogihara
****/
#ifndef __COVAL_H__
#define __COVAL_H__

#define __USE_MUTEX__	0
#if !defined(NULL)
#define  NULL	(void *)0
#endif
#define USE_LONGLONG	1
#define USE_LONGDOUBLE	0

#define COVAL_PROC_SUCCESS	0	// callback is done successfully
#define COVAL_PROC_IGNORED	1	// callback does no practical work
#define COVAL_PROC_ERROR	(-1)	// failure

struct _coval_proto;
typedef int (*coval_callback)(struct _coval_proto *, void *);

typedef struct _coval_center_info {
	unsigned short _count;     // reference counter of center coval
	unsigned char  _properties;
	unsigned char  _;   // reserved
} coval_center_info;

typedef struct _coval_proto {
	void *_p;  // if NULL, not bound
	union {
		struct _coval_proto *_cent; // link from member to cnter
		coval_center_info _cinfo;   // info. of center coval
	} _b;
	unsigned char _callself;
	coval_callback _cb;
	void *_ag; // argument for callback func
	struct _coval_proto *_nx; // link to next if _cb != NULL
	struct _coval_proto *_pv; // link to prev. if _cb != NULL
#if __USE_MUTEX__
	mutex	_mtx;
#endif
	long _d; // dummy for data
} coval_base;
typedef coval_base *coval_ref;

#define CovalInitialData(...)  \
	{ NULL, NULL, 0, NULL, NULL, NULL, NULL, /*NULL*/ __VA_ARGS__ }
#define CovalInit(Cp)   (Cp)->_p = NULL, (Cp)->_callself = 0, (Cp)->_cb = 0

#if __USE_MUTEX__
#define CovalTypeOf(TYP)	struct { TYP *_p; \
	union { coval_base *_cent; coval_center_info _cinfo; } _b; \
	unsigned char _callself; \
	coval_callback _cb; void *_ag; coval_base *_nx, *_pv; \
	mutex	_mtx; \
	TYP _d; }
#else
#define CovalTypeOf(TYP)	struct { TYP *_p; \
	union { coval_base *_cent; coval_center_info _cinfo; } _b; \
	unsigned char _callself; \
	coval_callback _cb; void *_ag; coval_base *_nx, *_pv; \
	TYP _d; }
#endif

#define IsCenter(Cp)	((Cp)->_p == &((Cp)->_d))

#define _rawBytesEq(A, B)  ((A)->_r[0] == (B)->_r[0] && (A)->_r[1] == (B)->_r[1])

void _coval_call(coval_base *);
#define CovalValue(Cp)	( (Cp)->_p ? *((Cp)->_p) : (Cp)->_d )
#define CovalMemberValue(Cp, M) ((Cp)->_p ? ((Cp)->_p)->M : (Cp)->_d.M)
#define CovalAssign(Cp, V)	if ( (Cp)->_p ) { \
	*((Cp)->_p) = (V); _coval_call((coval_base *)(Cp)); \
	}else (Cp)->_d = (V)
#define CovalAssignMember(Cp, M, V)	if ( (Cp)->_p ) { \
	(Cp)->_p->M = (V); _coval_call((coval_base *)(Cp)); \
	}else (Cp)->_d.M = (V)
#define CovalAssignMembers2(Cp, M, V, M2, V2)	if ( (Cp)->_p ) { \
	(Cp)->_p->M = (V); (Cp)->_p->M2 = (V2); \
	_coval_call((coval_base *)(Cp)); \
	}else (Cp)->_d.M = (V), (Cp)->_d.M2 = (V2)
#define CovalAssignMembers3(Cp, M, V, M2, V2, M3, V3)	if ( (Cp)->_p ) { \
	(Cp)->_p->M = (V); (Cp)->_p->M2 = (V2); (Cp)->_p->M3 = (V3); \
	_coval_call((coval_base *)(Cp)); \
	}else (Cp)->_d.M = (V), \
	(Cp)->_d.M2 = (V2), (Cp)->_d.M3 = (V3)
#define CovalAssignMembers4(Cp, M, V, M2, V2, M3, V3, M4, V4)	\
	if ( (Cp)->_p ) { \
	(Cp)->_p->M = (V); (Cp)->_p->M2 = (V2); (Cp)->_p->M3 = (V3); \
	(Cp)->_p->M4 = (V4); _coval_call((coval_base *)(Cp)); \
	}else (Cp)->_d.M = (V), (Cp)->_d.M2 = (V2), \
		(Cp)->_d.M3 = (V3), (Cp)->_d.M4 = (V4)

void _set_coval_function(coval_base *cov, coval_callback fn, void *arg);
#define SetCovalCallback(Cp, F, A)	 if ((Cp)->_p == NULL) { \
	(Cp)->_cb = (coval_callback)(F);  (Cp)->_ag = (A); } else \
	_set_coval_function((coval_base *)(Cp), (coval_callback)(F), (A))

#define SetCovalCallSelf(Cp, F)  (Cp)->_callself = (F)

void _bindCoval(coval_base *target, coval_base *cov, void *pp);
#define BindCoval(TARGp, Cp)	_bindCoval((coval_base *)(TARGp), (coval_base *)(Cp), NULL)

void _unbindCoval(coval_base *);
#define UnbindCoval(CV)	if ((CV) && !IsCenter(CV)) { \
	(CV)->_d = *(CV)->_p; _unbindCoval((coval_base *)CV); } else 0

#define BIND_FLAG_BREAKABLE	0x01	// quit if COVAL_PROC_IGNORED
void _setBindingProperty(coval_base *, unsigned int, int);
#define SetBindingLoopBreakable(Cp, F) \
	_setBindingProperty((coval_base *)(Cp), BIND_FLAG_BREAKABLE, (F))

/* Coval Bridge */
typedef int (*coval_bridge_convert)(void *to, const void *frm);
	/* Conversion function */

#define _CONV_FromMASK	0x03
#define _CONV_ToMASK	0x0c
#define _CONV_OpeMASK	0xf0
#define Check_mask	0x0f
#define Func_mask	_CONV_OpeMASK // 0xf0
#define _CONV_ToShift	2
#define BR_NONE		0
#define BR_INT		1
#define BR_UINT		2
#define BR_FLT		3
#define BR_IntToInt	(BR_INT | (BR_INT << _CONV_ToShift))
#define BR_IntToFloat	(BR_INT | (BR_FLT << _CONV_ToShift))
#define BR_UIntToInt	(BR_UINT | (BR_INT << _CONV_ToShift))
#define BR_UIntToFloat	(BR_UINT | (BR_FLT << _CONV_ToShift))
#define BR_FloatToInt	(BR_FLT | (BR_INT << _CONV_ToShift))
#define BR_FloatToFloat	(BR_FLT | (BR_FLT << _CONV_ToShift))
#define _BR_Conv	0x10
#define BR_NumNeg	0x20
#define BR_LogicNeg	0x30 // "to" should be int
#define BR_CopyBytes	0x40

typedef struct _bridge_base {
	coval_base *_cv[2];
	unsigned char _sw[2];	// check | Func/Ope.
	unsigned short _sz[2];	// size of data
	union { coval_bridge_convert func; unsigned char conv[2]; } _fn[2];
	coval_base bank[2]; 
} bridge_base;
	// conv[2]: from-type, to-type

#define CovalBridgeOf(TYP,TYQ)	struct { coval_base *_cv[2]; \
	unsigned char _sw[2]; unsigned short _sz[2]; \
	union { coval_bridge_convert func; unsigned char conv[2]; } _fn[2]; \
	CovalTypeOf(TYP) bank1; CovalTypeOf(TYQ) bank2; }
void _coval_bridge_initialize(bridge_base *, unsigned, unsigned);
#define CovalBridgeInit(TYP, TYQ, BRp) do { \
	typedef CovalBridgeOf(TYP,TYQ) _ty; _ty *p = (_ty *)(BRp); \
	p->_cv[0] = (coval_base *)&p->bank1; p->_cv[1] = (coval_base *)&p->bank2; \
	_coval_bridge_initialize((bridge_base *)(BRp), sizeof(TYP), sizeof(TYQ)); \
	}while(0)

#define BridgeCoval(BRp, X, TARGp)  _bindCoval((coval_base *)(TARGp), \
	((bridge_base *)(BRp))->_cv[X], NULL)

#define BridgeCovalMember(BRp, X, TARGp, M)	_bindCoval( \
	(coval_base *)(TARGp), \
	((bridge_base *)(BRp))->_cv[X], &(TARGp)->_p->M)

#define UnbridgeCoval(BRp, X)	UnbindCoval((BRp)->_cv[X])

int _setCovalBridgeConversion(bridge_base *brg, int idx, int conv, int check);
#define SetCovalBridgeConversion(BRp, X, CONV, CK) \
    _setCovalBridgeConversion((bridge_base *)(BRp), (X), (CONV), (CK))
#define SetCovalBridgeFunction(BRp, X, FN) \
    (BRp)->_sw[X] = Func_mask, (BRp)->_fn[X].func = (coval_bridge_convert)(FN)

#endif /* __COVAL_H__ */
